"""implement Martel parsers

The classes in this module are used by other Martel modules and not
typically by external users.

There are two major parsers, 'Parser' and 'RecordParser.'  The first
is the standard one, which parses the file as one string in memory
then generates the SAX events.  The other reads a record at a time
using a RecordReader and generates events after each read.  The
generated event callbacks are identical.

At some level, both parsers use "_do_callback" to convert mxTextTools
tags into SAX events.

XXX finish this documentation

XXX need a better way to get closer to the likely error position when
parsing.

XXX need to implement Locator

XXX won't the byte counts be wrong for MS Windows when the 'Parser'
reads both "\r\n" while the 'RecordParser', via readline, sees it as
just a '\n'?

"""

import urllib, pprint, traceback, sys, string
from xml.sax import xmlreader, _exceptions, handler

try:
    from mx import TextTools
except ImportError:
    import TextTools

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

# These exceptions are liable to change in the future
class ParserException(_exceptions.SAXException):
    """used when a parse cannot be done"""
    pass

class ParserPositionException(ParserException):
    def __init__(self, pos):
        ParserException.__init__(self,
                    "error parsing at or beyond character %d" % pos,
                                     None)
        self.pos = pos
    def __iadd__(self, offset):
        self.pos += offset
        self._msg = "error parsing at or beyond character %d" % self.pos
        return self

class ParserRecordException(ParserException):
    """used by the RecordParser when it can't read a record"""
    pass


# Uses a hack to support back references in mxTextTools!

# THIS MEANS SINGLE THREADED SUPPORT for anything using
# backreferences!  There is a much more complicated solution where the
# first element of any taglist is defined to contain the _match_group
# for that parse session.  I don't want to do that, since another
# approach is to modify mxTextTools to pass around an extra state
# object, or to write my own code.  (Did someone say NIH syndrome? :)
_match_group = {}



# The SAX startElements take an AttributeList as the second argument.
# Martel's attributes are always empty, so make a simple class which
# doesn't do anything and which I can guarantee won't be modified.
class MartelAttributeList(xmlreader.AttributesImpl):
    def getLength(self):
        return 0
    def getName(self, i):
        raise IndexError, i
    def getType(self, i):
        raise IndexError, i
    def getValue(self, i):
        raise IndexError, i
    def __len__(self):
        return 0
    def __getitem__(self, key):
        if type(key) == type(0):
            raise IndexError, key
        else:
            raise KeyError, key
    def keys(self):
        return []
    def values(self):
        return []
    def items(self):
        return []
    def has_key(self, key):
        return 0
    def get(self, key, alternative):
        return alternative

# singleton object shared amoung all startElement calls
_attribute_list = MartelAttributeList([])


def _do_callback(s, begin, end, taglist, cont_handler):
    """internal function to convert the tagtable into ContentHandler events

    's' is the input text
    'begin' is the current position in the text
    'end' is 1 past the last position of the text allowed to be parsed
    'taglist' is the tag list from mxTextTools.parse
    'cont_handler' is the SAX ContentHandler
    """
    for item in taglist:
        tag, l, r, subtags = item
        # If the tag's beginning is after the current position, then
        # the text from here to the tag's beginning are characters()
        if begin < l:
            cont_handler.characters(s[begin:l])
        else:
            # Some integrity checking
            assert begin == l, "begin = %d and l = %d" % (begin, l)

        # Named groups doesn't create ">ignore" tags, so pass them on
        # to the ContentHandler.  Unnamed groups still need a name so
        # mxTextTools can create subtags for them.  I named them
        # ">ignore" - dont' create events for them.
        if tag != ">ignore":
            cont_handler.startElement(tag, _attribute_list)

        # Recurse if it has any children
        if subtags:
            _do_callback(s, l, r, subtags, cont_handler)
        else:
            cont_handler.characters(s[l:r])
        begin = r

        if tag != ">ignore":
            cont_handler.endElement(tag)

    # anything after the last tag and before the end of the current
    # range are characters
    if begin < end:
        cont_handler.characters(s[begin:end])

def _parse_elements(s, tagtable, cont_handler, debug_level):
    """parse the string with the tagtable and send the ContentHandler events

    Specifically, it sends the startElement, endElement and characters
    events but not startDocument and endDocument.
    """
    if debug_level:
        import Generate
        Generate._position = 0

    result, taglist, pos = TextTools.tag(s, tagtable, 0, len(s))

    # Special case text for the base ContentHandler since I know that
    # object does nothing and I want to test the method call overhead.
    if cont_handler.__class__ != handler.ContentHandler:
        # Send any tags to the client (there can be some even if there
        _do_callback(s, 0, pos, taglist, cont_handler)

    if not result:
        if debug_level:
            return ParserPositionException(Generate._position)
        else:
            return ParserPositionException(pos)
    elif pos != len(s):
        return pos
    else:
        return None

# This needs an interface like the standard XML parser
class Parser(xmlreader.XMLReader):
    """Parse the input data all in memory"""

    def __init__(self, tagtable, (want_groupref_names, debug_level) = (0, 1)):
        xmlreader.XMLReader.__init__(self)

        assert type(tagtable) == type( () ), "mxTextTools only allows a tuple tagtable"
        self.tagtable = tagtable

        # WARNING: This attribute is set directly by Generate - it bypasses
        # the value used in __init__.
        # Used to tell if the global "match_group" dict needs to be cleared.
        self.want_groupref_names = want_groupref_names

        self.debug_level = debug_level
        
    def __str__(self):
        x = StringIO()
        pprint.pprint(self.tagtable, x)
        return x.getvalue()

    def parseFile(self, fileobj):
        """parse using the input file object

        XXX will be removed with the switch to Python 2.0, where parse()
        takes an 'InputSource'
        """
        # Just parse as a string
        self.parseString(fileobj.read())
    
    def parse(self, systemId):
        """parse using the URL"""
        # Just parse as a file ... which parses as a string
        self.parseFile(urllib.urlopen(systemId))
        
    def parseString(self, s):
        """parse using the given string

        XXX will be removed with the switch to Python 2.0, where parse()
        takes an 'InputSource'
        """
        self._cont_handler.startDocument()

        if self.want_groupref_names:
            _match_group.clear()

        # parse the text and send the SAX events
        result = _parse_elements(s, self.tagtable, self._cont_handler,
                                 self.debug_level)

        if result is None:
            # Successful parse
            pass

        elif isinstance(result, _exceptions.SAXException):
            # could not parse record, and wasn't EOF
            self._err_handler.fatalError(result)
        
        else:
            # Reached EOF
            pos = result
            self._err_handler.fatalError(ParserPositionException(pos))
        
        # Send an endDocument event even after errors
        self._cont_handler.endDocument()

    def close(self):
        pass

class RecordParser(xmlreader.XMLReader):
    """Parse the input data a record at a time"""
    def __init__(self, format_name, record_tagtable,
                 (want_groupref_names, debug_level),
                 make_reader, reader_args = ()):
        """parse the input data a record at a time

        format_name - XML tag name for the whole data file
        record_tagtable - mxTexTools tag table for each record
        want_groupref_names - flag to say if the match_group table needs to
              be reset (will disappear with better support from mxTextTools)

        make_reader - callable object which creates a RecordReader; first
              parameter will be an input file object
        reader_args - optional arguments to pass to make_reader after the
              input file object
        """
        xmlreader.XMLReader.__init__(self)
        
        self.format_name = format_name
        assert type(record_tagtable) == type( () ), \
               "mxTextTools only allows a tuple tagtable"
        self.tagtable = record_tagtable
        self.want_groupref_names = want_groupref_names
        self.debug_level = debug_level
        self.make_reader = make_reader
        self.reader_args = reader_args
    
    def __str__(self):
        x = StringIO()
        pprint.pprint(self.tagtable, x)
        return "parse records: " + x.getvalue()
    
    def parseFile(self, fileobj):
        """parse using the input file object

        XXX will be removed with the switch to Python 2.0, where parse()
        takes an 'InputSource'
        """
        self._cont_handler.startDocument()
        
        try:
            reader = apply(self.make_reader, (fileobj,) + self.reader_args)
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            # something unexpected happened
            # so call it a fatal error and stop
            outfile = StringIO()
            traceback.print_exc(file=outfile)
            self._err_handler.fatalError(ParserRecordException(
                outfile.getvalue(), sys.exc_info()[1]))
            self._cont_handler.endDocument()
            return

        if self.want_groupref_names:
            _match_group.clear()
        
        self._cont_handler.startElement(self.format_name, _attribute_list)
        filepos = 0  # can get mixed up with DOS style "\r\n"
        while 1:
            try:
                record = reader.next()  
            except (KeyboardInterrupt, SystemExit):
                raise
            except:
                # something unexpected happened (couldn't find a record?)
                # so call it a fatal error and stop
                outfile = StringIO()
                traceback.print_exc(file=outfile)
                self._err_handler.fatalError(ParserRecordException(
                    outfile.getvalue(), sys.exc_info()[1]))
                self._cont_handler.endDocument()
                return
            
            if record is None:
                break
            result = _parse_elements(record, self.tagtable, self._cont_handler,
                                     self.debug_level)

            if result is None:
                # Successfully read the record
                pass
            elif isinstance(result, _exceptions.SAXException):
                # Wrong format or a SAX problem, but this is recoverable
                self._err_handler.error(result)
            else:
                # Did not reach end of string, but this is recoverable
                pos = filepos + result
                self._err_handler.error(ParserPositionException(pos))

            filepos = filepos + len(record)

        self._cont_handler.endElement(self.format_name)
        self._cont_handler.endDocument()

    def parse(self, systemId):
        """parse using the URL"""
        # Just parse it as a file
        self.parseFile(urllib.urlopen(systemId))
        
    def parseString(self, s):
        """parse using the given string

        XXX will be removed with the switch to Python 2.0, where parse()
        takes an 'InputSource'
        """
        # Just parse it as a file
        strfile = StringIO(s)
        self.parseFile(strfile)

    def close(self):
        pass

# This is entirely too complex, but I don't see a way to simplify it
class HeaderFooterParser(xmlreader.XMLReader):
    """Header followed by 0 or more records followed by a footer"""
    def __init__(self, format_name,
                 make_header_reader, header_reader_args, header_tagtable,
                 make_reader, reader_args, record_tagtable,
                 make_footer_reader, footer_reader_args, footer_tagtable,
                 (want_groupref_names, debug_level)):
        xmlreader.XMLReader.__init__(self)

        self.format_name = format_name

        self.make_header_reader = make_header_reader
        self.header_reader_args = header_reader_args
        self.header_tagtable = header_tagtable

        self.make_reader = make_reader
        self.reader_args = reader_args
        self.record_tagtable = record_tagtable
        
        self.make_footer_reader = make_footer_reader
        self.footer_reader_args = footer_reader_args
        self.footer_tagtable = footer_tagtable
        
        self.want_groupref_names = want_groupref_names
        self.debug_level = debug_level

    def __str__(self):
        x = StringIO()
        pprint.pprint( (self.header_tagtable, self.record_tagtable,
                        self.footer_tagtable), x)
        return "header footer records: " + x.getvalue()

    def parseString(self, s):
        strfile = StringIO(s)
        self.parseFile(strfile)

    def parse(self, systemID):
        self.parseFile(urllib.urlopen(systemID))

    def parseFile(self, fileobj):
        self._cont_handler.startDocument()
        self._cont_handler.startElement(self.format_name, _attribute_list)

        if self.want_groupref_names:
            _match_group.clear()

        # Read the header
        if self.make_header_reader is not None:
            try:
                header_reader = apply(self.make_header_reader,
                                      (fileobj,) + self.header_reader_args)
                header = header_reader.next()
            except (KeyboardInterrupt, SystemExit):
                raise
            except:
                # something unexpected happened
                # so call it a fatal error and stop
                outfile = StringIO()
                traceback.print_exc(file=outfile)
                self._err_handler.fatalError(ParserRecordException(
                    outfile.getvalue(), sys.exc_info()[1]))
                self._cont_handler.endDocument()
                return

            # Parse the text (if any) and send the SAX events
            if header is None:
                header = ""

            result = _parse_elements(header, self.header_tagtable,
                                     self._cont_handler, self.debug_level)

            if result is None:
                # Successful parse
                pass
            elif isinstance(result, _exceptions.SAXException):
                # could not parse header, and wasn't EOF
                self._err_handler.fatalError(result)
                self._cont_handler.endDocument()
                return
            else:
                # Reached EOF
                pos = result
                self._err_handler.fatalError(ParserPositionException(pos))
                self._cont_handler.endDocument()
                return

        # We've successfully parsed the header, now parse the records

        # Get any forward data from the header reader
        if self.make_header_reader is None:
            x, lines = fileobj, []
        else:
            x, lines = header_reader.remainder()

        # Make the record reader
        try:
            reader = apply(self.make_reader, (fileobj,) + self.reader_args,
                           {"lines": lines})
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            # something unexpected happened
            # so call it a fatal error and stop
            outfile = StringIO()
            traceback.print_exc(file=outfile)
            self._err_handler.fatalError(ParserRecordException(
                outfile.getvalue(), sys.exc_info()[1]))
            self._cont_handler.endDocument()
            return

        # filepos can get mixed up with DOS style "\r\n"
        if header is None:
            filepos = 0
        else:
            filepos = len(header)
        
        while 1:
            try:
                record = reader.next()
##                print "RECORD"
##                print repr(record)
            except (KeyboardInterrupt, SystemExit):
                raise
            except:
                # Something strange happened, so perhaps it's a footer?
                if self.make_footer_reader is None:
                    # No footer possible, so call it a fatal error and stop
                    outfile = StringIO()
                    traceback.print_exc(file=outfile)
                    self._err_handler.fatalError(ParserRecordException(
                        outfile.getvalue(), sys.exc_info()[1]))
                    self._cont_handler.endDocument()
                    return
                record = None
                break

            if record is None:
                # Reached EOF, but need to see if an empty footer is okay
                break

            result = _parse_elements(record, self.record_tagtable,
                                     self._cont_handler, self.debug_level)

            if result is None:
                # Successfully parsed the record
                pass
            else:
                # Failed to parse the record, so maybe it's a footer?
                if self.make_footer_reader is not None:
                    # Perhaps, so have to try
                    break
                else:
                    # No footer possible
                    if isinstance(result, _exceptions.SAXException):
                        # Wrong format or a SAX problem, but recoverable
                        if isinstance(result, ParserPositionException):
                            # fix offset if possible
                            result += filepos
                        self._err_handler.error(result)
                    else:
                        # Did not reach end of string, but recoverable
                        pos = filepos + result
                        self._err_handler.error(ParserPositionException(pos))

            filepos = filepos + len(record)

        # At this point one of:
        #  - Reached EOF, so record is None and reader.lines == []
        #  - The record could not be read, so record is None and
        #        reader.lines may contain data
        #  - The record could not be parsed, so record contains text and
        #        reader.lines may contain data

        x, lines = reader.remainder()
        if record is not None:
            # For the case when the record footer and the format have
            # the same RecordReader
            lines = string.split(record, "\n") + lines

        if self.make_footer_reader is not None:
            # Make the footer reader
            try:
                reader = apply(self.make_footer_reader,
                               (fileobj,) + self.footer_reader_args,
                               {"lines": lines})
                footer = reader.next()
            except (KeyboardInterrupt, SystemExit):
                raise
            except:
                # something unexpected happened
                # so call it a fatal error and stop
                outfile = StringIO()
                traceback.print_exc(file=outfile)
                self._err_handler.fatalError(ParserRecordException(
                    outfile.getvalue(), sys.exc_info()[1]))
                self._cont_handler.endDocument()
                return

            if footer is None:
                footer = ""

            result = _parse_elements(footer, self.footer_tagtable,
                                     self._cont_handler, self.debug_level)

            if result is None:
                # Successful parse
                pass
            elif isinstance(result, _exceptions.SAXException):
                if isinstance(result, ParserPositionException):
                    result += filepos+ result.pos
                self._err_handler.fatalError(result)
                self._cont_handler.endDocument()
                return
            else:
                # Reached EOF
                pos = filepos + result
                self._err_handler.fatalError(ParserPositionException(pos))
                self._cont_handler.endDocument()
                return

            # see if there are any lines left over
            x, lines = reader.remainder()
            
        if lines or fileobj.readline():
            pos = filepos + len(footer)
            self._err_handler.fatalError(ParserPositionException(pos))
            self._cont_handler.endDocument()
            return

        # Hey, it finished.
        self._cont_handler.endElement(self.format_name)
        self._cont_handler.endDocument()

