# The existing parsers are in-memory.  For large data files, like
# swissprot, that requires too much memory.

# On the other hand, the records aren't all that large (there's just a
# lot of them.)  So this will scan the input for each start record
# indicator, up to and including the end record indicator, then pass
# the results to the SAX handler.

# Depends on a file being only a set of records, and records must
# terminated by a unique string (like "//\n").

import string

SIZEHINT = 100000

# pre-Python 2.0 cStringIO doesn't have a "readlines" so fake it.
# NB: StringIO doesn't allow a sizehint for readlines (also fixed in 2.0)
class FakeReadLines:
    def __init__(self, readline):
        self.readline = readline
    def readlines(self, sizehint):
        line = self.readline()
        if not line:
            return []
        return [line]

class RecordReader:
    def __init__(self, infile):
        self.infile = infile
    def next(self):
        raise NotImplementedError
    def remaining(self):
        raise NotImplementedError
    
class StartsWith(RecordReader):
    def __init__(self, infile, str, sizehint = SIZEHINT, lines = []):
        RecordReader.__init__(self, infile)
        self.str = str
        self.sizehint = sizehint

        if not sizehint or not hasattr(self.infile, "readlines"):
            # Force reading a line at a time, either because readlines
            # isn't available or because of the request to limit
            # read-ahead.
            self.readlines = FakeReadLines(self.infile.readline).readlines
        else:
            self.readlines = self.infile.readlines
        self.lines = lines[:] + self.readlines(sizehint)
            
        # Guaranteed to have read at least one line, if not at EOF
        if self.lines:
            # XXX use "startsWith"
            assert self.lines[0][:len(str)] == str, \
                   "first line not in the right format (need %s in %s)" % \
                   (repr(str), repr(self.lines[0]))

    def next(self):
        # Will always have a line pending, or no lines for EOF
        lines = self.lines
        if not lines:
            return None

        # There's a line pending, and we've already verified that it's
        # the beginning characters of the record.
        readlines = self.readlines
        start_lineno = 1
        test_lineno = 1
        str = self.str
        n = len(str)

        while 1:
            for test_lineno in range(start_lineno, len(lines)):
                if lines[test_lineno][:n] == str:
                    s = string.join(lines[:test_lineno], "")
                    self.lines = lines[test_lineno:]
                    return s

            # terminator not found, so read more lines
            start_lineno = test_lineno
            new_lines = readlines(self.sizehint)
            if not new_lines:
                # Reached EOF, so return the record
                # (remember, guaranteed to have at least one line)
                s = string.join(lines, "")
                self.lines = []
                return s
            lines.extend(new_lines)

    def remainder(self):
        return self.infile, self.lines

class EndsWith(RecordReader):
    """Read a record at a time, where the record ends with a specified line.

    This is only usable with line oriented files. The final line of the
    record must be the same as the given string, including the newline.
    
    BUG: Could change it so the first part of the last line matches a
    given string.  Not needed for now, and much cleaner in Python 2.0.
    """
    def __init__(self, infile, terminator, sizehint = SIZEHINT, lines = []):
        # XXX Should fix - startsWith?
        assert terminator[-1] == "\n", \
               "only accepts complete lines as a terminator"
        RecordReader.__init__(self, infile)
        self.terminator = terminator
        self.lines = lines[:]
        self.sizehint = sizehint

        if hasattr(self.infile, "readlines"):
            self.readlines = self.infile.readlines
        else:
            self.readlines = FakeReadLines(self.infile.readline).readlines
        
    def next(self):
        lines = self.lines
        terminator = self.terminator
        start_lineno = 0
        test_lineno = 0
        readlines = self.readlines
        
        while 1:
            for test_lineno in range(start_lineno, len(lines)):
                if lines[test_lineno] == terminator:
                    s = string.join(lines[:test_lineno+1], "")
                    self.lines = lines[test_lineno+1:]
                    return s

            # terminator not found, so read more lines
            start_lineno = test_lineno
            new_lines = readlines(self.sizehint)
            if not new_lines:
                if not lines:
                    return None  # end of file
                else:
                    raise AssertionError("cannot find terminator: %s" % \
                                         repr(self.terminator))
            lines.extend(new_lines)


    def remainder(self):
        return self.infile, self.lines

class Until(RecordReader):
    """Read up to but not including a line starting with the given text"""
    def __init__(self, infile, str, sizehint = SIZEHINT, lines = []):
        RecordReader.__init__(self, infile)
        self.str = str
        self.lines = lines[:]
        self.sizehint = sizehint

        if hasattr(self.infile, "readlines"):
            self.readlines = self.infile.readlines
        else:
            self.readlines = FakeReadLines(self.infile.readline).readlines

    def next(self):
        lines = self.lines
        str = self.str
        start_lineno = 0
        test_lineno = 0
        readlines = self.readlines
        
        while 1:
            for test_lineno in range(start_lineno, len(lines)):
                if lines[test_lineno].startswith(str) and test_lineno != 0:
                    s = string.join(lines[:test_lineno], "")
                    self.lines = lines[test_lineno:]
                    return s
            # terminator not found, so read more lines
            start_lineno = test_lineno
            new_lines = readlines(self.sizehint)

            if not new_lines:
                if not lines:
                    return None  # end of file
                elif lines[test_lineno].startswith(str) and test_lineno == 0:
                    s = string.join(lines, "")
                    self.lines = []
                    return s
                else:
                    print repr(new_lines)
                    print repr(lines)
                    raise AssertionError("cannot find terminator: %s" % \
                                         repr(str))

            lines.extend(new_lines)

    def remainder(self):
        return self.infile, self.lines

class CountLines(RecordReader):
    """Read a specified (fixed) number of lines"""
    def __init__(self, infile, count, sizehint = SIZEHINT, lines = []):
        assert count >= 0
        RecordReader.__init__(self, infile)
        self.lines = lines[:]
        self.count = count

    def next(self):
        lines = self.lines

        for i in range(len(lines), self.count):
            line = self.infile.readline()
            if not line:
                if not lines:
                    return None
                raise EOFError, "need %d lines of input (have %d)" % \
                      (self.count - i, len(lines))
            lines.append(line)
            
        x = lines[:self.count]
        self.lines = lines[self.count:]
        return x

    def remainder(self):
        return self.infile, self.lines


class Nothing(RecordReader):
    """Reads nothing"""
    def __init__(self, infile, sizehint = SIZEHINT, lines = []):
        RecordReader.__init__(self, infile)
        self.lines = lines[:]

    def next(self):
        return None
    
    def remainder(self):
        return self.infile, self.lines

class Everything(RecordReader):
    """Reads everything"""
    def __init__(self, infile, sizehint = SIZEHINT, lines = []):
        RecordReader.__init__(self, infile)
        self.sizehint = sizehint
        self.lines = lines[:]

        if not sizehint or not hasattr(self.infile, "readlines"):
            self.readlines = FakeReadLines(self.infile.readline).readlines
        else:
            self.readlines = infile.readlines

    def next(self):
        lines = self.lines
        while 1:
            new_lines = self.readlines(self.sizehint)
            if not new_lines:
                break
            lines.extend(new_lines)

        if not lines:
            return None
        
        s = string.join(lines, "")
        self.lines = []
        return s

    def remainder(self):
        return self.infile, self.lines
